/*
  -------------------- file driver.c
  Main driver for ntuple search.
  Created 16 Nov 93 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "our_common.h"
#include "driver.h"

#include "prog/running.h"
#include "prog/timer.h"
#include "libinclude/ntup.h"

int getrusage P((int, struct rusage *));

void
driver(
FILE *f,                              /* our output stream */
const struct to_ntup_struct *findto,  /* input data */
const struct aux_struct * aux,        /* auxiliary input information */
const int alg,                        /* algorithm to use; ALG_XX */
const char *title                     /* title for summary (NULL okay) */
)
{
  struct from_ntup_struct findfrom;
  struct rusage   rbegin, rend;

  job_start(f);   /* counts business below, is probably negligible */
  getrusage(RUSAGE_SELF, &rbegin);

  if (alg == ALG_TABU)
    /* ntup_find_tabu(findto, & findfrom, aux->tabu_max_iterations); */
		fatal("tabu search not distributed with this release.");
  else if (alg == ALG_BACK1)
    ntup_find_back1(findto, & findfrom,
            aux->back_action,aux->back_print,aux->back_max_nodes);
  else if (alg == ALG_BACK2)
    ntup_find_back2(findto, & findfrom,
            aux->back_action,aux->back_print,aux->back_max_nodes,
            aux->back2_timeslice_nodes, aux->back2_restrict_width);
  else if (alg == ALG_WIDE1)
    /* ntup_find_wide1(findto, & findfrom, aux->wide1_max_width); */
		fatal("wide1 search not distributed with this release.");
  else
    fatal("ntuple_driver: unknown or unimplemented algorithm requested.");

  if (findfrom.status == NTUP_OK) {
    int i;
    const int numtups = findfrom.tuples_found;

    fprintf(f, "\n");
    fprintf(f, "Tuple finding procedure completed normally.");
    fprintf(f, "  Presenting %d solution tuples.\n", numtups);
    fprintf(f, "\n");
    for (i = 0; i < numtups; i++) {
      const struct tuple_struct *tup = & findfrom.tuples[i];
      if (numtups == 1)   fprintf(f, "Tuple");
      else                fprintf(f, "Tuple %d", i);
      fprintf(f, " (satisfies p=%d, q=%d):\n", tup->p, tup->q);
      ntup_print(f, tup->pixels, tup->n, 1);         /* 1 = use long format */
      fprintf(f, "\n");
    }
  }
  else {
    fprintf(f,"\n");
    fprintf(f, "--------------------\n");
    fprintf(f, "--------------------\n");
    fprintf(f, "Error reported by tuple-finding procedure:\n");
    ntup_print_error(f,& findfrom);
    fprintf(f, "\n");         /* ntup_print_error does not print newline */
  }


  /* summary line */
  {
    long u_cpu_sec, u_cpu_msec;

    getrusage(RUSAGE_SELF, &rend);
    end_timing2(&rbegin.ru_utime, &rend.ru_utime, &u_cpu_sec, &u_cpu_msec);

    fprintf(f, "\n");
    fprintf(f,"Summary line:\n");
    fprintf(f,"# %-25s ucpu=%5ld.%02ld  p,q:",
      title ? title : "untitled", u_cpu_sec, u_cpu_msec/10000);
    if (findfrom.status == NTUP_OK) {
      const int numtups = findfrom.tuples_found;
      int i;
      if (numtups == 0)
        fprintf(f," none");
      else {
        for (i = 0; i < numtups; i++) {
          const struct tuple_struct *tup = & findfrom.tuples[i];
          fprintf(f,"%3d,%d", tup->p, tup->q);
        }
      }
    }
    else
      fprintf(f, " (ERROR)\n");
    fprintf(f,"\n");
  }

  if (findfrom.status == NTUP_OK)
    ntup_from_free(& findfrom);

  job_end(f);
}
