/* -------------------- file main.c
  Created 16 Nov 93 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <sys/time.h>
#include <limits.h>

#include "our_common.h"
#include "driver.h"

#include "prog/argproc.h"
#include "prog/bitarray.h"
#include "prog/fileio.h"
#include "prog/getnum.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"
#include "prog/random.h"
#include "prog/util.h"

#include "libinclude/ntup.h"

#define DEFAULT_SEED            0
#define DEFAULT_BACKGROUND      WHITE

static int  algorithm;    /* ALG_XX from driver.h */
static char *title;


/* not static because compiler gives warning if static */
const char *copyright = "Copyright 1993 by Andrew Shapira.";
const char *Progname  = "nt";

extern int printf();


void                  /* argproc routines refer to this */
usage(
const char  *errmess
)
{
  /* use pound sign so it shows up in test suite summary files */
  if (errmess)
    fprintf(stderr, "### Error in command line arguments.\n\n");
  else
    fprintf(stderr, "### Displaying usage message (-help specified).\n");
  fprintf(stderr, "----- %s usage: (output goes to standard output)\n", Progname);
  fprintf(stderr,
    "  [-help]  -n n  (-b | -w | -bw)  [(-p p -q q) | -nomaxp]  [-pi r c]\n");
  fprintf(stderr,
    "  (-back1 N | -back2 N T W )\n");
  fprintf(stderr,
    "  [-randseed | -seed seed]  [-count | -one]  [-bpnone -bpfirst -bpall -bpcanon] \n");
  fprintf(stderr,
    "  -pos x f1 f2 .. fx  -neg y g1 g2 .. gy  [-notrim]  [-pbg b]  [-nbg B]\n");
  fprintf(stderr,
    "  (-postlist F1 F2 .. Fx | -postkind (pos | pi))  (-negtlist ..)\n");
  fprintf(stderr,
    "  [-ioflags f]  [-u update_interval]  [-title t]\n");
  fprintf(stderr, "\n");
  fprintf(stderr,
    "-help                display usage message and exit\n");
  fprintf(stderr,
    "-n n                 find tuples that have n pixels (must have n >= 2)\n");
  fprintf(stderr,
    "-b, -w, -bw          black-only, white-only, black and white tuples\n");
  fprintf(stderr,
    "-p p -q q            search only specified p,q (def=pseudo-critical p,q)\n");
  fprintf(stderr,
    "-nomaxp              don't use 'maximize p' feature internally in back1 or back2\n");
  fprintf(stderr,
    "-pi r c              pi has r rows and c cols (def=each min. pos. dimension)\n");
  fprintf(stderr,
    "-back1 N             simple backtrack; stop for given p,q after N nodes (0=inf)\n");
  fprintf(stderr,
    "-back2 N T W         complex backtrack; stop for given p,q after N nodes (0=inf)\n");
  fprintf(stderr,
    "                     timeslice for T nodes, restrict width to W (0=inf)\n");
  fprintf(stderr,
    "-randseed / -seed s  use random/specified random num. seed (def seed is %d)\n", DEFAULT_SEED);
  fprintf(stderr,
    "-count               back1,back2: enumerate solutions, for encountered p,q\n");
  fprintf(stderr,
    "-one                 back1,back2: stop after one solution, for encountered p,q\n");
  fprintf(stderr,
    "                     (Note: for -count and -one, node limit takes priority.)\n");
  fprintf(stderr,
    "-bpnone              back1 and back2 print no solutions (def)\n");
  fprintf(stderr,
    "-bpfirst             back1 and back2 print only first solution encountered\n");
  fprintf(stderr,
    "-bpall               back1 and back2 print all solutions encountered\n");
  fprintf(stderr,
    "-bpcanon             back1 and back2 print only canonical solutions encountered\n");
  fprintf(stderr,
    "-pos x f1 f2 ..      pixels for x positive chars are in files f1, f2, ...\n");
  fprintf(stderr,
    "-postlist F1 F2 ..   translation lists for pos. chars are in files F1, F2 ..\n");
  fprintf(stderr,
    "-postkind (pos | pi)         trans lists for pos. chars: pos=pos. offsets; pi=pi\n");
  fprintf(stderr,
    "-neg, -negtlist, -negtkind   (like 'pos' options)\n");
  fprintf(stderr,
    "-notrim              do not trim characters (def=trim)\n");
  fprintf(stderr,
    "-pbg b / -nbg B      positive/negative background color is b/B (def=%d)\n",
      DEFAULT_BACKGROUND);
  fprintf(stderr,
    "-ioflags f           io flags.  Masks: %d=report input  %d=report results\n",
      IOF_REPORT_INPUT, IOF_REPORT_RESULTS);
  fprintf(stderr,
    "                       %d=no time-dependent output\n",
      IOF_NO_TIME_DEPENDENT_OUTPUT);
  fprintf(stderr,
    "-u update_interval   s. between updates, or: def=%d=never %d=strategic %d=often\n",
      UPDATE_NONE,UPDATE_STRATEGIC,UPDATE_OFTEN);
  fprintf(stderr,
    "-title t             display title 't' on summary line\n");
  fprintf(stderr,
    "\n");
  fprintf(stderr,
    "Color legend: %d=white  %d=black  %d=nomatch  %d=match\n",
      WHITE,BLACK,PSEUDOCOLOR_NOMATCH,PSEUDOCOLOR_MATCH);
  if (errmess) {
    fprintf(stderr,"\n");
    fatal(errmess);
  }
  exit(2);
}

static void
readpixels(
FILE *f,
const char *kind,               /* positive or negative (for diagnostics) */
const int i,                    /* ith char of this kind (for diagnostics) */
struct character_struct *dest,  /* where to put character */
void  *(*allocfunc) P((const int))
)
{
  int min,max;

  if (! mat_uchar_read2_with_dim
                        (f,& dest->a,&dest->nrows,&dest->ncols,allocfunc)) {
    fprintf(stderr, "Error in reading %s input character %d.\n", kind, i+1);
    fatal(NULL);
  }
  mat_uchar_minmax2(dest->a,dest->nrows,dest->ncols, &min,&max);
  if ((min < PSEUDOCOLOR_MIN) || (max > PSEUDOCOLOR_MAX)) {
    fprintf(stderr, "Error in reading %s input character %d.\n", kind, i+1);
    fprintf(stderr, "Pixel values must be in range %d .. %d\n",
      PSEUDOCOLOR_MIN, PSEUDOCOLOR_MAX);
    fatal("Illegal pixel value(s).");
  }
}

static void
readtrans(
FILE *f,
const char *kind,               /* positive or negative (for diagnostics) */
const int i,                    /* ith char of this kind (for diagnostics) */
struct character_struct *dest,  /* where to put character */
void  *(*allocfunc) P((const int))
)
{
  int ntrans;
  int d;

  if (! getnum(f, &ntrans)) {
    fprintf(stderr,
      "Error in reading translations for %s input character %d.\n", kind, i+1);
    fatal(NULL);
  }
  dest->ntrans = ntrans;
  dest->trans_list = (*allocfunc) (ntrans * sizeof(dest->trans_list[0]));
  for (d = 0; d < ntrans; d++) {
    int dr,dc;
    if ((! getnum(f, &dr)) || (! getnum(f, &dc))) {
      fprintf(stderr,
        "Error in reading translations for %s input character %d.\n",kind,i+1);
      fatal(NULL);
    }
    dest->trans_list[d].dr = dr;
    dest->trans_list[d].dc = dc;
  }
}

/*  Set up translations to cover pi, positive offsets only.
*/
static void
trans_plus(
struct character_struct *chars,
const int nchars,
void  *(*allocfunc) P((const int))
)
{
  int i;
  for (i = 0; i < nchars; i++) {
    struct character_struct * const h = & chars[i];
    ntup_trans_make_rectangular(h,0,h->nrows-1,0,h->ncols-1,allocfunc);
  }
}

/*  Set up translations to cover pi.
*/
static void
trans_pi(
const int pi_rows,
const int pi_cols,
struct character_struct *chars,
const int nchars,
void  *(*allocfunc) P((const int))
)
{
  int i;
  for (i = 0; i < nchars; i++) {
    struct character_struct * const h = & chars[i];
    ntup_trans_make_rectangular
      (h,1-pi_rows,h->nrows-1,1-pi_cols,h->ncols-1,allocfunc);
  }
}


/* returns 1 if any actual trimming occurred, 0 if nothing changed */
static int
trimarr(
struct character_struct *chars,
const int nchars,
const int background
)
{
  int i;
  unsigned char **newa;
  int newrows,newcols;
  int trimmed;

  trimmed = 0;
  for (i = 0; i < nchars; i++) {
    mat_uchar_trim2(chars[i].a,chars[i].nrows,
      chars[i].ncols,background,&newa,&newrows,&newcols,xmalloc);
    if ((newrows != chars[i].nrows) || (newcols != chars[i].ncols))
      trimmed = 1;
    mat_uchar_free(chars[i].a, xfree);
    chars[i].a = newa;
    chars[i].nrows = newrows;
    chars[i].ncols = newcols;
  }
  return(trimmed);
}

/* returns 1 if any actual trimming occurred, 0 if nothing changed */
static int
trim(
struct to_ntup_struct *findto
)
{
  int trimmed;

  /* assumes chars of findto were allocated using xmalloc */
  trimmed = trimarr(findto->pos.chars, findto->pos.nchars, findto->pos.bg);
  if (trimarr(findto->neg.chars, findto->neg.nchars, findto->neg.bg))
    trimmed = 1;
  return(trimmed);
}

static void
proc_args(
struct to_ntup_struct *const findto,
struct aux_struct * const aux,
const int argc,
char      **argv
)
{
  const TKIND_POS = 0;
  const TKIND_PI = 1;
  int   an;
  int   done_n,done_b,done_w,done_bw,done_p,done_q,done_pi,done_pos,done_neg;
  int   done_postlist, done_negtlist, done_postkind, done_negtkind;
  int   done_seed,done_randseed,done_u,done_help,done_alg,done_title;
  int   done_notrim,done_count,done_one;
  int   done_bpnone,done_bpfirst,done_bpall,done_bpcanon;
  int   done_ioflags,done_pbg,done_nbg,done_nomaxp;
  int   n,p,q,seed,u,back_max_nodes,pos_tkind,neg_tkind;
  int   back2_timeslice_nodes, back2_restrict_width;
  int   npos,nneg,ioflags,pbg,nbg;
  int   pi_rows, pi_cols;
  int   tmpvec[3];

  extern pid_t  getpid P(( void ));
  extern long   srandom P(( int ));

  done_n = done_b = done_w = done_bw = done_p = done_q = done_pi = 0;
  done_pos = done_neg = done_seed = done_randseed = done_u = 0;
  done_postlist = done_negtlist = done_postkind = done_negtkind = 0;
  done_help = done_alg = done_title = done_notrim = 0;
  done_count = done_one = done_ioflags = 0;
  done_bpnone = done_bpfirst = done_bpall = done_bpcanon = 0;
  done_pbg = done_nbg = done_nomaxp = 0;

  DO_GCCHAPPY(back2_timeslice_nodes = back2_restrict_width = 0;)
  DO_GCCHAPPY(pi_rows = pi_cols = 0;)
  DO_GCCHAPPY(pos_tkind = neg_tkind = 0;)

  for (an = 1; an < argc; ) {
    if (argnull("-help",argc,argv,&an,&done_help)) {}
    else if (argint("-n",1,INT_MAX,argc,argv,&an,&n,&done_n)) {}
    else if (argnull("-b",argc,argv,&an,&done_b)) {}
    else if (argnull("-w",argc,argv,&an,&done_w)) {}
    else if (argnull("-bw",argc,argv,&an,&done_bw)) {}
    else if (argint("-p",1,INT_MAX,argc,argv,&an,&p,&done_p)) {}
    else if (argint("-q",1,INT_MAX,argc,argv,&an,&q,&done_q)) {}
    else if (argnull("-nomaxp",argc,argv,&an,&done_nomaxp)) {}
    else if (argint_vec("-pi",2,1,INT_MAX,argc,argv,&an,tmpvec,&done_pi)) {
      pi_rows = tmpvec[0];
      pi_cols = tmpvec[1];
    }
    else if (argint("-back1",0,INT_MAX,argc,argv,&an,&back_max_nodes,&done_alg))
      algorithm = ALG_BACK1;
    else if (argint_vec("-back2",3,0,INT_MAX,argc,argv,&an,tmpvec,&done_alg)) {
      algorithm = ALG_BACK2;
      back_max_nodes = tmpvec[0];
      back2_timeslice_nodes = tmpvec[1];
      back2_restrict_width = tmpvec[2];
    }
    else if (argnull("-randseed",argc,argv,&an,&done_randseed)) {}
    else if (argint("-seed",INT_MIN,INT_MAX,argc,argv,&an,&seed,&done_seed)) {}
    else if (argnull("-count",argc,argv,&an,&done_count)) {}
    else if (argnull("-one",argc,argv,&an,&done_one)) {}
    else if (argnull("-bpnone",argc,argv,&an,&done_bpnone)) {}
    else if (argnull("-bpfirst",argc,argv,&an,&done_bpfirst)) {}
    else if (argnull("-bpall",argc,argv,&an,&done_bpall)) {}
    else if (argnull("-bpcanon",argc,argv,&an,&done_bpcanon)) {}
    else if (argint("-pos",1,INT_MAX,argc,argv,&an,&npos,&done_pos)) {
      /* cloned below for -neg */
      int i;
      findto->pos.nchars = npos;
      findto->pos.chars = xmalloc(npos * sizeof(findto->pos.chars[0]));
      for (i = 0; i < npos; i++) {
        FILE *f;
        if (an >= argc)  usage("Missing part of '-pos' arg.");
        f = xfopen(argv[an++],"r");
        readpixels(f,"positive",i,&findto->pos.chars[i],xmalloc);
        xfclose(f);
      }
    }
    else if (argnull("-postlist",argc,argv,&an,&done_postlist)) {
      /* cloned below for -negtlist */
      int i;
      if (! done_pos)  usage("Must specify -pos before -postlist.");
      if (done_postkind)  usage("Cannot specify both -postkind and -postlist.");
      for (i = 0; i < npos; i++) {
        FILE *f;
        if (an >= argc)  usage("Missing part of '-postlist' arg.");
        f = xfopen(argv[an++],"r");
        readtrans(f,"positive",i,&findto->pos.chars[i],xmalloc);
        xfclose(f);
      }
    }
    else if (argnull("-postkind",argc,argv,&an,&done_postkind)) {
      /* cloned below for -negtkind */
      if (done_postlist)  usage("Cannot specify both -postkind and -postlist.");
      if (argnull("pos",argc,argv,&an,NULL))
        pos_tkind = TKIND_POS;
      else if (argnull("pi",argc,argv,&an,NULL))
        pos_tkind = TKIND_PI;
      else
        usage("Missing or illegal option to -postkind.");
    }
    else if (argint("-neg",1,INT_MAX,argc,argv,&an,&nneg,&done_neg)) {
      /* cloned above for -pos*/
      int i;
      findto->neg.nchars = nneg;
      findto->neg.chars = xmalloc(nneg * sizeof(findto->neg.chars[0]));
      for (i = 0; i < nneg; i++) {
        FILE *f;
        if (an >= argc)  usage("Missing part of '-neg' arg.");
        f = xfopen(argv[an++],"r");
        readpixels(f,"negative",i,&findto->neg.chars[i],xmalloc);
        xfclose(f);
      }
    }
    else if (argnull("-negtlist",argc,argv,&an,&done_negtlist)) {
      /* cloned above for -postlist */
      int i;
      if (! done_neg)  usage("Must specify -neg before -negtlist.");
      if (done_negtkind)  usage("Cannot specify both -negtkind and -postlist.");
      for (i = 0; i < nneg; i++) {
        FILE *f;
        if (an >= argc)  usage("Missing part of '-negtlist' arg.");
        f = xfopen(argv[an++],"r");
        readtrans(f,"negative",i,&findto->neg.chars[i],xmalloc);
        xfclose(f);
      }
    }
    else if (argnull("-negtkind",argc,argv,&an,&done_negtkind)) {
      /* cloned above for -postkind */
      if (done_negtlist)  usage("Cannot specify both -negtkind and -negtlist.");
      if (argnull("pos",argc,argv,&an,NULL))
        neg_tkind = TKIND_POS;
      else if (argnull("pi",argc,argv,&an,NULL))
        neg_tkind = TKIND_PI;
      else
        usage("Missing or illegal option to -negtkind.");
    }
    else if (argnull("-notrim",argc,argv,&an,&done_notrim)) {}
    else if (argint("-pbg",0,PSEUDOCOLOR_MAX,argc,argv,&an,&pbg,&done_pbg)) {}
    else if (argint("-nbg",0,PSEUDOCOLOR_MAX,argc,argv,&an,&nbg,&done_nbg)) {}
    else if (argint("-ioflags",0,INT_MAX,argc,argv,&an,&ioflags,&done_ioflags))
      {}
    else if (argint("-u",UPDATE_MIN,INT_MAX,argc,argv,&an,&u,&done_u)) {}
    else if (argstr("-title",argc,argv,&an,&title,&done_title)) {}

    else {
      fprintf(stderr, "'%s'\n", argv[an]);
      usage("Unrecognized argument.");
    }
  }

  if (done_help)   usage(NULL);
  if (! done_alg)  usage("Must specify -back1, or -back2.");
  if (! done_n)    usage("Must give -n option.");
  if (! done_pos)  usage("Must specify -pos.");
  if (! done_neg)  usage("Must specify -neg.");
  if ((! done_b) && (! done_w) && (! done_bw))
    usage("Must specify -b, -w, or -bw.");
  if ((done_b && done_w) || (done_w && done_bw) || (done_b && done_bw))
    usage("Must specify only one of -b, -w, or -bw.");
  if (done_seed && done_randseed)
    usage("Cannot specify both -seed and -randseed.");
  if ((done_p) && (! done_q))  usage("Must give -q if -p is given.");
  if ((! done_p) && (done_q))  usage("Must give -p if -q is given.");
  
  findto->n = n;

  if (done_nomaxp && (done_p || done_q))
    usage("Cannot specify -nomaxp with -p or -q.");
  if (done_p)
    {  findto->pq_kind = PQ_SPECIFIC;  findto->p = p;  findto->q = q;  }
  else if (done_nomaxp)
    findto->pq_kind = PQ_QTHENP_NOMAXP;
  else
    findto->pq_kind = PQ_QTHENP;

  if (done_b)         findto->tuple_kind = TUPLE_BLACKONLY;
  else if (done_w)    findto->tuple_kind = TUPLE_WHITEONLY;
  else if (done_bw)   findto->tuple_kind = TUPLE_BLACKWHITE;


  if ((algorithm == ALG_BACK1) || (algorithm == ALG_BACK2)) {
    if (done_count && done_one)
      usage("Cannot specify both -count and -one.");
    if (done_count)     aux->back_action = BACK_ACTION_COUNT;
    else if (done_one)  aux->back_action = BACK_ACTION_ONE;
    else
      usage("With -back1 and -back2, must specify -count or -one.");
    /* Note: aux->back_max_nodes and aux->back2_timeslice_nodes
       are doubles.
    */
    aux->back_max_nodes = back_max_nodes;
    if (algorithm == ALG_BACK2) {
      aux->back2_timeslice_nodes = back2_timeslice_nodes;
      aux->back2_restrict_width = back2_restrict_width;
    }
    if (done_bpnone+done_bpfirst+done_bpall+done_bpcanon > 1)
      usage("Must specify at most one of -bpnone,-bpfirst,-bpall,-bpcanon.");
    if (done_bpnone)
      aux->back_print = BACK_PRINT_NONE;
    else if (done_bpfirst)
      aux->back_print = BACK_PRINT_FIRST;
    else if (done_bpall)
      aux->back_print = BACK_PRINT_ALL;
    else if (done_bpcanon)
      aux->back_print = BACK_PRINT_CANON_ONLY;
    else
      aux->back_print = BACK_PRINT_NONE;   /* default */
  }
  else {
    if (done_count || done_one || done_bpnone || done_bpfirst || done_bpall || done_bpcanon)
      usage("Use -count,-one,-bpnone,-bpfirst,-bpall,-bpcanon only with -back1 or -back2");
  }

  if (done_randseed) {
    printf("Using random seed.");
    seed = environ_rand();
  }
  else if (done_seed)
    printf("Using specified seed.");
  else {
    printf("Using default seed.");
    seed = DEFAULT_SEED;
  }
  printf("  Seed is %d.\n", seed);
  (void) srandom(seed);

  if (done_pbg) findto->pos.bg = pbg;
  else          findto->pos.bg = DEFAULT_BACKGROUND;

  if (done_nbg) findto->neg.bg = nbg;
  else          findto->neg.bg = DEFAULT_BACKGROUND;
  
  if (done_u)  findto->update_interval = u;
  else         findto->update_interval = UPDATE_NONE;

  if (! done_title)  title = NULL;

  if (done_ioflags)  findto->ioflags = ioflags;
  else               findto->ioflags = 0;

  if (! done_notrim) {
    if (trim(findto))
      printf("Input character(s) trimmed.\n");
    else
      printf("Input characters do not need trimming.\n");
  }
  else {
    printf("No trimming attempted (-notrim specified).\n");
  }

  if (done_pi) {
    findto->pi_rows = pi_rows;
    findto->pi_cols = pi_cols;
  }
  else {
    /*  Find bounding box of pi if it was not specified.
        (Do this after trimming.)
    */
    int i;

    findto->pi_rows = findto->pi_cols = INT_MAX;
    for (i = 0; i < npos; i++) {
      const int rows = findto->pos.chars[i].nrows;
      const int cols = findto->pos.chars[i].ncols;
      if (rows < findto->pi_rows)  findto->pi_rows = rows;
      if (cols < findto->pi_cols)  findto->pi_cols = cols;
    }
  }

  /* Do this translation business after pi_rows is set.
  */

  if (done_postkind) {
    if (pos_tkind == TKIND_POS)
      trans_plus(findto->pos.chars, findto->pos.nchars, xmalloc);
    else if (pos_tkind == TKIND_PI)
      trans_pi(findto->pi_rows,findto->pi_cols,
        findto->pos.chars, findto->pos.nchars, xmalloc);
  }
  else if (done_postlist) {}
  else
    usage("Must specify -postkind or -postlist.");

  if (done_negtkind) {
    if (neg_tkind == TKIND_POS)
      trans_plus(findto->neg.chars, findto->neg.nchars, xmalloc);
    else if (neg_tkind == TKIND_PI)
      trans_pi(findto->pi_rows,findto->pi_cols,
        findto->neg.chars, findto->neg.nchars, xmalloc);
  }
  else if (done_negtlist) {}
  else
    usage("Must specify -negtkind or -negtlist.");
}

static void
run(
FILE *outf,
struct to_ntup_struct *findto,
struct aux_struct *aux
)
{
  fprintf(outf,"Running with ");
#ifdef  ASSERTIONS_ON
  fprintf(outf,"ASSERTIONS_ON,");
#else
  fprintf(outf,"no ASSERTIONS_ON,");
#endif
#ifdef  HEAVY_ASSERTIONS_ON
  fprintf(outf," HEAVY_ASSERTIONS_ON.\n");
#else
  fprintf(outf," no HEAVY_ASSERTIONS_ON.\n");
#endif
  {
    extern int region_assertions_on;
    if (region_assertions_on)
      fprintf(outf, "   back2/region.c: HEAVY_ASSERTIONS_ON\n");
  }
  fprintf(outf,"\n");
  fprintf(outf,"\n");

  findto->f             = outf;
  driver(outf,findto,aux,algorithm,title);
}


int
main(
const int argc,
char **argv
)
{
  struct to_ntup_struct findto;
  struct aux_struct aux;

  setvbuf(stdout,NULL,_IOLBF,0);    /* line buffer stdout */
  setvbuf(stderr,NULL,_IOLBF,0);    /* line buffer stderr */

  printf("Ntuplib version " NTUPLIB_VERSION);
  printf(", Copyright (C) 1993 Andrew Shapira.\n");
	printf("\n");
	printf("Ntuplib comes with ABSOLUTELY NO WARRANTY; for details see the copy of\n");
	printf("the GNU General Public License that is distributed with the Ntuplib source\n");
	printf("code.  Ntuplib is free software, and you are welcome to redistribute it\n");
	printf("under certain conditions; for details see the copy of the GNU General\n");
	printf("Public License that is distributed with the Ntuplib source code.\n");
	printf("\n");

  bit_global_init();

  proc_args(&findto,&aux,argc,argv);
  run(stdout,&findto,&aux);
  /* NOTE: some of the fields of findto are allocated, and not freed.  */

  return(0);
}

#ifdef OLD
The old inner loop of main.c/run().  Problems are specified
by having all the characters in one file.

{
    bsbegin();

    if (! getnum(inf,&findto->npos))  break;
    if (findto->npos < 1)
      fatal("Input error: number of positive characters.");
    if (! getnum(inf,&findto->nneg))
      fatal("Input error: number of negative characters.");
    if (findto->nneg < 1)
      fatal("Input error: number of negative characters.");

    findto->pos = bsalloc(sizeof(findto->pos[0]) * findto->npos);
    findto->neg = bsalloc(sizeof(findto->neg[0]) * findto->nneg);
    for (i = 0; i < findto->npos; i++)
      readpixels(stdin,"positive",i,& findto->pos[i],bsalloc);
    for (i = 0; i < findto->nneg; i++)
      readpixels(stdin,"negative",i,& findto->neg[i],bsalloc);

    ntuple_driver(outf,findto,algorithm);

    bsend();
  }
#endif /*OLD*/
