/*
  -------------------- file bitword.c
  Created 9 May 92 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Support for bit manipulation routines.

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*  USERS MUST CALL bitword_global_init() before calling these routines.
*/

#define SRCFILE_LIBPROG_BITWORD

#include "our_common.h"
#include "bitword.h"
#include "memalloc.h"

/* MACHINE DEPENDENT */
/* bitword_rmask[i] = mask with smallest i bits set */
BITWORD bitword_rmask[] = {
  0,
  0x00000001, 0x00000003, 0x00000007, 0x0000000f,
  0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
  0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
  0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
  0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
  0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
  0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
  0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff
};


/*  Initialize tables used by bit manipulation functions.  Users can call
    this function multiple times with the expectation that this function
    and its descendants will do almost nothing, if this function has
    been called previously.
*/
void
bitword_global_init()
{
  bitword_count_table_init();
  bitword_small_table_init();
  bitword_large_table_init();
}



/* ----------------------------------- bit small masks
*/

/*  Generate an array of masks 0x1, 0x3, 0x7, 0xf, 0x1f, etc.
    ret[i] == small i+1 bits are set.
*/
void
bitword_make_rmask(
BITWORD     *ret,
const int   len
)
{
  int   i;

  if (len == 0)
    return;
  ret[0] = (BITWORD) 0x1;
  for (i = 1; i < len; i++)
    ret[i] = (ret[i-1] << 1) | 1;
}


/* ----------------------------------- bit counting
*/

/*  more than just these constants must be changed to change things */
#define CTABLE_BITS   16
#define CTABLE_MAX    65536   /* 2^CTABLE_BITS-1 */

unsigned char *_bitword_count_table = NULL;   /* count of how many bits set */

/*  Allocate memory for use in bitword_count() using xmalloc().
    Initialize _bitword_count_table.  Most users do not need to call this
    routine directly; that is taken care of automatically by
    bitword_global_init().
*/
void
bitword_count_table_init(void)
{
  BITWORD   i;
  int       j;
  int       count;

  if (! _bitword_count_table) {
    _bitword_count_table = xmalloc(CTABLE_MAX);
    for (i = 0; i < CTABLE_MAX; i++) {
      for (count = 0, j = 0; j < CTABLE_BITS; j++) {
        if (BITWORD_IS_BIT_SET(i,j))
          count++;
      }
      _bitword_count_table[i] = count;
    }
  }
}

void
bitword_count_table_free(void)
{
  xfree(_bitword_count_table);
  _bitword_count_table = NULL;
}

/*  more than just these constants must be changed to change things */
#define BTABLE_BITS   16
#define BTABLE_MAX    65536   /* 2^BTABLE_BITS */

/* ----------------------------------- finding of smallest bits
*/


char *_bitword_small_table = NULL;/* smallest bit in for each bit pattern */

/*  Allocate memory for use in bitword_find_smallest() using xmalloc().
    Initialize _bitword_small_table.  Most users do not need to call this
    routine directly; that is taken care of automatically by
    bitword_global_init().
*/
void
bitword_small_table_init(void)
{
  BITWORD   i;
  int       j;

  if (! _bitword_small_table) {
    _bitword_small_table = xmalloc(BTABLE_MAX);
    _bitword_small_table[0] = -1;
    for (i = 1; i < BTABLE_MAX; i++) {
      for (j = 0; j < BTABLE_BITS; j++) {
        if (BITWORD_IS_BIT_SET(i, j)) {
          _bitword_small_table[i] = j;
          break;
        }
      }
    }
  }
}

void
bitword_small_table_free(void)
{
  xfree(_bitword_small_table);
  _bitword_small_table = NULL;
}

/* ----------------------------------- finding of largest bits
*/

char *_bitword_large_table = NULL;/* largest bit in for each bit pattern */

/*  Allocate memory for use in bitword_find_largest() using xmalloc().
    Initialize _bitword_large_table.  Most users do not need to call this
    routine directly; that is taken care of automatically by
    bitword_global_init().
*/
void
bitword_large_table_init(void)
{
  BITWORD   i;
  int       j;

  if (! _bitword_large_table) {
    _bitword_large_table = xmalloc(BTABLE_MAX);
    _bitword_large_table[0] = -1;
    for (i = 1; i < BTABLE_MAX; i++) {
      for (j = BTABLE_BITS-1; j >= 0; j--) {
        if (BITWORD_IS_BIT_SET(i, j)) {
          _bitword_large_table[i] = j;
          break;
        }
      }
    }
  }
}

void
bitword_large_table_free(void)
{
  xfree(_bitword_large_table);
  _bitword_large_table = NULL;
}

/* ----------------------------------- bit reversal
  TODO - initialization interface is inconsistent with above routines
       - figure out whether this should be rewritten / simplified
*/

/* _table[i] == smallest bits of i reversed */
BITWORD *_bitword_rev_table;
int _bitword_rev_len;         /* number of bits rev_table is set up for */
int _bitword_rev_len_m_16;    /* bitword_rev_len-16 (only valid if len>16) */
int _bitword_32_m_rev_len;    /* 32-_bitword_rev_len (only valid if len>16) */

void
bitword_rev_table_alloc(
const int   b,                            /* number of bits */
void  *(*allocfunc) P((const int))
)
{
  _bitword_rev_table = (*allocfunc) ( sizeof(BITWORD) * (1 << MIN(b,16)) );
  _bitword_rev_len = b;
  if (b > 16) {
    _bitword_rev_len_m_16 = b-16;
    _bitword_32_m_rev_len = 32-b;
  }
}

void
bitword_rev_table_init(void)
{
  const int len = MIN(_bitword_rev_len,16);
  const BITWORD max = 1 << len;
  BITWORD i,w;
  int j;

  for (i = 0; i < max; i++) {
    for (j = 0, w = 0; j < len; j++) {
      if (BITWORD_IS_BIT_SET(i,j))
        BITWORD_SET(w,len-1-j);
    }
    _bitword_rev_table[i] = w;
    D(printf("rev[%08x]=%08x\n", i, w);)
  }
}

void
bitword_rev_table_free(
void  (*freefunc) P((void *))
)
{
  (*freefunc) (_bitword_rev_table);
}
