/*
  -------------------- file io.c

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <ctype.h>

#include "our_common.h"
#include "bitword.h"
#include "io.h"
#include "memalloc.h"
#include "util.h"

/*  Calculate the number of character positions required to display an int.
    MACHINE DEPENDENT.

    COULDDO: get rid of the division, in a machine independent way.
*/
int
fieldlen(
int v
)
{
  if (v < 0)
    return(1+fieldlen(-1 * v));   /* MACHINE DEPENDENT */
  if (v <= 9)
    return(1);
  return(1+fieldlen(v/10));
}

/*  Print n spaces.  If n < 1, no spaces are printed.  It is acceptable to
    pass negative values for n.
*/
void
spaces_print(
FILE *f,
int n
)
{
  static char *spaces;
  static int  alloc_size = 0;
  int         i;

  if (n < 1)
    return;
  if (n+1 > alloc_size) {
    if (alloc_size)
      xfree(spaces);
    alloc_size = MAX(64, n*2);
    spaces = xmalloc(alloc_size);
    for (i = 0; i < alloc_size; i++)
      spaces[i] = ' ';
  }
  spaces[n] = '\0';
  fputs(spaces, f);
  spaces[n] = ' ';
}

/*
  It is possible to use this routine recursively.
  Only the outermost routine is allowed to call this one with
  colret == NULL.

  Does not end the last line with a newline (the caller must do this).

  The "printfunc" must know what output stream is being used if
  the stream is not stdout.

*/
int
list_print(
FILE          *f,
const void    * const base,       /* const void * */
const int     elementwidth,
const int     n,
const int     depth,        /* Note: at outermost level if depth==0 */
const int     innermost,    /* at innermost level? */
const int     incol,
const int     leftmargin,   /* lowest possible value is 1 (is 1-based) */
const int     rightmargin,
const int     inneedwidth,  /* need at least this many to print an object */
void          *printinfo,
int           (*printfunc) P((const void *, const void *, const int)),
int           (*comparfunc) P((const void *, const void *, const void *))
)
{
  int   i;
  int   howmany;
  int   col = incol;
  const int continuemargin = leftmargin+2*(depth+1);
  const int needwidth = inneedwidth+2*(depth+1);
  /* const int outermost = (depth == 0); */
  int   newline;

  if (col < leftmargin) {
    spaces_print(f, leftmargin-col);
    col = leftmargin;
  }
  fprintf(f, "< ");
  col += 2;
  howmany = 1;
  for (i = 0; i < n; i++) {
#ifdef  DEBUG
    fprintf(f, "\n%%");
    (void) (*printfunc)(  (const void *) (((const char *) base)+i*elementwidth),
                  printinfo, col);
    fprintf(f, "%%\n");
#endif  DEBUG
    newline = 0;
    /*  Gather duplicates; for example, print  x ^ 5 rather than x x x x x. */
    if (i > 0) {
      /* this guy is different than the previous one */
      if ((*comparfunc)(
                (const void *) (((const char *) base)+i*elementwidth),
                (const void *) (((const char *) base)+(i-1)*elementwidth),
                printinfo)) {
D(printf("different from previous; howmany=%d\n", howmany);)
        if (howmany > 1) {
          fprintf(f, "^%d", howmany);
          col += fieldlen(howmany)+1;
          howmany = 1;
        }
        /* if ((! innermost) && (i < n-1)) {}{}{}{} */
        if (! innermost) {
          fprintf(f, "\n");
          spaces_print(f, continuemargin-1);
          col = continuemargin;
          newline = 1;
        }
      }
      /* this guy is the same as the previous one */
      else {
D(printf("same as previous\n");)
        howmany++;
        if (i == n-1) {
          fprintf(f, "^%d", howmany);
          col += fieldlen(howmany)+1;
        }
        continue;
      }
    }

    /*  We are going to print something.  */
    if (col > rightmargin-3-needwidth) {
      fprintf(f, "\n");
      spaces_print(f, continuemargin-1);
      col = continuemargin;
      newline = 1;
    }
    if ((! newline) && (i > 0)) {
      fprintf(f, " ");
      col ++;
    }
    col = (*printfunc)( (const void *) (((const char *) base)+i*elementwidth),
                  printinfo, col);
  }
  fprintf(f, " >");
  col += 2;
  return(col);
}

/*  Routines for printing and comparing various kinds of objects */

int
list_int_print(
const void  *element,       /* const int * */
const void  *printinfo,     /* int[1]: field width */
const int   col
)
{
  printf("%*d", (( const int *) printinfo) [0], *((const int *) element));
  return(col + ((const int *) printinfo) [0]);
}

int
list_u_print(
const void  *element,       /* const unsigned int * */
const void  *printinfo,     /* int[1]: field width */
const int   col
)
{
  printf("%*d",
    (( const int *) printinfo) [0], *((const unsigned int *) element));
  return(col + ((const int *) printinfo) [0]);
}

int
list_uchar_print(
const void  *element,       /* const unsigned char * */
const void  *printinfo,     /* int[1]: field width */
const int   col
)
{
  printf("%*d",
    (( const int *) printinfo) [0], *((const unsigned char *) element));
  return(col + ((const int *) printinfo) [0]);
}

int
list_int_compare(
const void *p1,     /* const int * */
const void *p2,     /* const int * */
const void *info    /* ignored */
)
{
  return(*(const int *)p1 - *(const int *)p2);
}

int
list_u_compare(
const void *p1,     /* const unsigned int * */
const void *p2,     /* const int * */
const void *info    /* ignored */
)
{
  return(*(const unsigned int *)p1 - *(const unsigned int *)p2);
}
  

int
list_uchar_compare(
const void *p1,     /* const unsigned char * */
const void *p2,     /* const unsigned char * */
const void *info    /* ignored */
)
{
  return(*(const unsigned char *)p1 - *(const unsigned char *)p2);
}

/*  Print the bits of an unsigned int from left to right.  This routine
    is not efficient; a faster routine could be made by using a lookup
    table for pieces of the word.

    Does not print trailing newline.
*/
void
bit_u_print(
FILE        *f,
unsigned int w
)
{
  int   i;

  for (i = 0; i < sizeof(unsigned int) * 8; i++) {
    if (BITWORD_IS_BIT_SET(w, (sizeof(unsigned int) * 8 - 1) - i))
      fprintf(f, "1");
    else
      fprintf(f, "0");
  }
}


/* Front end to strtol().  Parses a string to make sure it can possibly
    be a number, then sends it to strtol().  It is intended that the numbers
    accepted by this routine are a subset (possibly proper, possibly not
    proper) of the strings that are handled by strtol().  Unlike strtol(),
    we do not allow garbage at the end of a string.
*/
void
our_strtol(
const char *s,      /* same as arg 1 to strtol() */
char **ptr,         /* same as arg 2 to strtol() */
const int base,     /* same as arg 3 to strtol() */
int   *status_ret,  /* 0 if s can be processed, 1 if error */
long  *val_ret      /* if *status_ret==1, this is strtol(s,ptr,base) */
)
{
  const char *p;
  extern long strtol P((const char *, char **, int));

  for (p = s; *p && isspace(*p); p++) {}
  if (*p == '-')
    p++;
  while (*p) {
    if ((*p < '0') || (*p > '9'))
      goto fail;
    p++;
  }
  *status_ret = 0;
  *val_ret = strtol(s, ptr, base);
  return;

  fail:
  *status_ret = 1;
  return;
}

void
our_atol(
const char *s,      /* same as arg 1 to strtol() */
int   *status_ret,  /* 1 if s can be processed, 0 if not */
long  *val_ret      /* if *status_ret==1, this is strtol(s,ptr,base) */
)
{
  our_strtol(s,NULL,10,status_ret,val_ret);
}

void
our_atoi(
const char *s,      /* same as arg 1 to strtol() */
int   *status_ret,  /* 1 if s can be processed, 0 if not */
int   *val_ret      /* if *status_ret==1, this is strtol(s,ptr,base) */
)
{
  long    val;

  our_strtol(s,NULL,10,status_ret,&val);
  if (! *status_ret)
    *val_ret = (int) val;
}

/*  This routine does essentially what "%b" (binary) would do in sprintf().
    Converts nonnegative integer "val" to a null-terminated string of unsigned
    chars, putting the result in "valstr".  Returns the length of the string
    (not counting the trailing '\0').  For val==0, the string is "0" and
    the length is 1.  If valstr==NULL, just returns the length.
    It is a fatal error to call this routine with val<0.
*/
int
itob(val, valstr)
const int   val;            /* value to be converted */
char        *valstr;        /* where to put returned string */
{
  unsigned int    uval;     /* unsigned version of "val" */
  unsigned int    mask;   
  int             bit;      /* value of this bit */
  int             found;    /* have we found any 1's left (from left)? */
  int             length;   /* how many bits so far */

  if (val < 0)
    fatal("Negative value in itob().");
  uval = val;
  found = 0;
  length = 0;
  mask = 1 << (8*sizeof(int) - 1);
  while (1) {
    bit = (uval & mask) ? 1 : 0;
    if (bit)
      found = 1;
    if (found) {
      if (valstr)
        valstr[length] = bit + '0';
      length++;
    }
    if (mask == 1)
      break;
    mask >>= 1;
  }
  if (valstr) {
    if (! found)
      valstr[length++] = '0';
    valstr[length] = '\0';
  }
  else
    if (! found)
      length++;
  return(length);
}

