/*
  -------------------- file random.c
  Created 14 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <time.h>
#include <setjmp.h>
#define SRCFILE_LIBPROG_RANDOM
#include "our_common.h"
#include "random.h"
#include "vector.h"

/*
  Generate a word based on environmental parameters such as the clock.
  This function is itself ANSI-C, but has some non-portable features:
      1.  It depends on vec_uchar_hash(), which assumes a 32-bit word size.

  The function vec_uchar_hash() computes a word of bits, given an
  array of unsigned chars.  (The function vec_u_hash() is similar,
  given an array of unsigned ints.)  We compute several words using
  various environmental quantities as input to the vec_xx_hash() functions.
  The words are then XOR-ed together to form the final result.
*/
unsigned
environ_rand(
void
)
{
  unsigned w1;    /* from tmpnam() */
  unsigned w2;    /* real time */
  unsigned w3;    /* setjmp */
  unsigned w4;    /* uninitialized memory from automatic allocation */
  unsigned w5;    /* uninitialized memory from malloc */


  /* uninitialized memory (do this first so that code below does not
     put memory into a fixed state)
  */
  {                                         /* w4 */
    const int BUFFLEN=64;
    char buff[BUFFLEN];

    w4 = vec_uchar_hash(buff,BUFFLEN);
  }
  {                                         /* w5 */
    const int BUFFLEN=64;
    char *const buff = malloc(BUFFLEN);

    if (! buff) fatal("environ_rand malloc");
    w5 = vec_uchar_hash(buff,BUFFLEN);
    (void) free(buff);
  }

  /* tmpnam() */
  {
    char * tmpnam P((char *));
    const unsigned char * const fname = tmpnam(NULL);
                                /* type cast from char to unsigned char */

    w1 = vec_uchar_hash(fname,strlen(fname));
  }

  /* real time */
  {
    time_t time P((time_t *));
    size_t strftime P((char *, size_t, const char *, const struct tm * ));

    time_t const timet_info = time(NULL);
    struct tm *const tm_ptr = localtime(&timet_info);
    const int BUFFLEN=64;
    char buff[BUFFLEN];

    /* %Y = year with century (1900-...) */
    /* %j = day within year.  (001-366) */
    /* %H = 24-hour (00-23) */
    /* %M = minute (00-59) */
    /* %S = second (00-61) */
    if (strftime(buff, BUFFLEN, "%Y%j%H%M%S", tm_ptr) == 0)
      fatal("environ_rand()");
    w2 = vec_uchar_hash((unsigned char *) buff,strlen(buff));
  }

  /* setjmp */
  {
    jmp_buf jdummy;
    const int len = sizeof(jdummy);
    union {
      jmp_buf j;
      char chars[len];
    } junion;

    (void) setjmp(junion.j);
    w3 = vec_uchar_hash((unsigned char *) junion.chars, len);
  }

  return(w1 ^ w2 ^ w3 ^ w4 ^ w5);
}
